;;; -*- Syntax: Ansi-Common-Lisp; Base: 10; Mode: lisp; Package: imho -*-
;;; $Id: form-elements.lisp,v 1.179 2003/05/13 22:05:57 craig Exp $
;;;
;;; Copyright (c) 1999 - 2003 onShore Development, Inc.
;;;
;;; See the file 'COPYING' in this directory for terms.

(in-package :imho)

;; ------------------------------------------------------------
;; framework class: html-form
;;
;; An html-form is any simple or complex component that accepts input.
;; html-form objects may be nested without limitation. The outside
;; object will provide an HTML <FORM> tag.

(defclass html-form (html-element method-caller)
  ((primitive-values :documentation "If this element is the outermost
form, a set of complete references to primitive-html-forms is stashed
here.")
   (form-state       :accessor form-state
                     :initform nil)))

;; html-form-primitive objects reflect INPUT tags in generated HTML,
;; such as text fields, checkboxes, and text areas, which have string
;; values returned in form submission.  The primitive-value slot
;; receives the unparsed value returned by a form POST.

(defclass html-form-primitive (html-form)
  ((primitive-value :accessor primitive-value
                    :initform nil)))

;; Track all primitive form elements in the outermost html-form, so
;; that they can receive values in a separate phase when the form is
;; posted (then we don't have to refer to the request anymore).

(defvar *form-rendering* nil)
    
(defmethod render-html :before ((self html-form-primitive) stream)
  (declare (ignore stream))
  (with-slots (element-external-name value)
    self
    (push (list self element-external-name value)
          (slot-value *form-rendering* 'primitive-values))))

;; There is a distinction between dynamic child elements (those which
;; are added by repeat-element), and static ones (defined by
;; defbindings).  Dynamic elements may have non-unique names, and are
;; replaced during every request cycle.  Static elements must have
;; unique internal names, and are persistent across requests.

(defmethod preawake :before ((self html-form))
  (when-bind (children (slot-value self 'children))
    (let ((cleaned-hash (make-hash-table)))
      (maphash (lambda (k v &aux newval)
                 (dolist (v v)
                   (unless (second v)
                     (push v newval)))
                 (setf (gethash k cleaned-hash) newval))
               children)
      (setf (slot-value self 'children) cleaned-hash))))

;; allow other around methods to ensure form context is setup.

;; fixme - typep called twice

(defmacro with-form-context ((self stream) &body body)
  `(with-slots (element-external-name form-state element-method element-frame)
    ,self
    (if (or *form-rendering* (not (typep ,self 'html-form)))

        (progn
          (when (typep ,self 'html-form)
            (setf form-state nil))
          ,@body)
        
        (let ((*form-rendering* ,self))
          (setf form-state nil)
          (setf (slot-value ,self 'primitive-values) nil)
          (html-stream
           ,stream
           ((:form :method :post
                   :enctype "multipart/form-data"
                   :target element-frame
                   :name element-external-name
                   :action (if element-method
                               (typecase element-method
                                 (symbol
                                  (build-element-url ,self :reference (dyn-refer-wm element-method nil)))
                                 (string
                                  element-method))
                               ""))
            (progn ,@body)))))))
  

;; The <FORM ...> tag is generated and *form-rendering* is set to t
;; when the first (enclosing) html-form element is found

(defmethod render-html :around ((self html-form) stream)
  (with-form-context (self stream)
    (call-next-method)))

(define-wm process-form-submit ((element html-form))
  (flet ((default-processor (element)
             (intern (concatenate 'string "PROCESS-" (symbol-name (type-of element)))
                     (symbol-package (type-of element)))))
    (let ((action nil))
      (block process
        ;; search children for subtypes of form-submit-element
        ;; retrieve the action of the first one found
        (mapc #'(lambda (elem)
                  (if (and (subtypep (type-of elem) 'html-form)
                           (element-value elem)
                           (maybecar (element-method elem)))
                      (progn
                        ;; (cmsg "Invoking ~A action" (element-internal-name elem))
                        (setq action (element-method elem))
                        (return-from process))))
              (reverse (cons element (element-children element)))))
      (apply (etypecase action
               (function (progn
                           (when *warn-deprecated*
                             (cmsg "WARNING: old-style Lambda found in form action"))
                           action))
               (null (let ((wm (lookup-wm (default-processor element))))
                       (if wm
                           (wmethod-body wm)
                           #'ignore)))
               (symbol (if (lookup-wm action)
                           (wmethod-body (lookup-wm action))
                           (progn
                             (when *warn-deprecated*
                               (cmsg "WARNING: old-style function found in form action")
                               (symbol-function action))))))
             (list element)))))


;; Look up the value of a form element in a request, by external
;; (HTML) name.

(defun get-form-value (name request)
  "returns the value of form element named NAME in REQUEST"
  (let ((values (mapcar #'cadr (remove-if-not (lambda (x)
                                                (equal name (car x)))
                                              (request-client-content request)))))
    (case (length values)
      (1
       (let ((value (car values)))
         (if (< 0 (length value))
             value
             nil)))
      (t
       values))))

;; Set the values of html-form-primitive elements using unparsed form
;; value.

(defun take-primitive-values (self request)
  (when (request-client-content request)
    (dolist (primitive-element (slot-value self 'primitive-values))
      (destructuring-bind (pelement element-name old-value)
          primitive-element
        (declare (ignore old-value))
        (let ((new-value (get-form-value element-name request)))
          ;; (cmsg "Set PVal of ~s to ~s" pelement new-value)
          (setf (primitive-value pelement) new-value))))
    t))

;; Descend the rendering tree, calling accept-input on each html-form
;; instance.

(defmethod take-values-from-request ((self html-form))
  (with-slots (children)
    self
    (when children
      (dolist (child-list (hashvalues children))
        (dolist (instance child-list)
          (take-values-from-request (car instance)))))
    (let ((value (accept-input self)))
      (setf (slot-value self 'value) value))))

;; Default is to ignore the value of a html-form element: subclasses
;; must implement this to assume a value other than that they already
;; possess.

(defmethod accept-input (self)
  (slot-value self 'value))

;; After accepting input, set the state of the form to :valid if it
;; hasn't been set to :invalid inside of accept-input.

(defmethod accept-input :around ((self html-form))
  (with-slots (form-state)
    self
    (unwind-protect
         (unless form-state
           (call-next-method))
      (unless (eql form-state :invalid)
        (setf form-state :valid)))))

;; ------------------------------------------------------------
;; html-element: auto-list
;;
;; If 'rows' is 1, this is drawn as a drop-down list; if > 1, as a
;; conventional list.
;; 
;; raw:
;; 
;; (("foo" #<x>)
;;  ("678768")
;;  ("bar" #<y>))
;; 
;; wcof-equal
;; car
;; 
;; (#<loc 1> #<loc 2>)
;; 
;; wcof-equal
;; name
;; 
;; (1 2 3 4 5 6)
;; 
;; eql
;; princ-to-string

(defclass auto-list (html-form-primitive)
  ((driver        :initarg :driver
                  :initform (make-list-driver nil)
                  :writer set-list-driver)
   (any           :initarg :any
                  :initform nil)
   (any-value     :initarg :any-value
                  :initform :any)
   (none          :initarg :none
                  :initform nil)
   (rows          :initarg :rows
                  :initform 1)
   (multiple      :initarg :multiple
                  :initform nil)
   (current-value :initform nil)
   (state         :initform nil)))

(defun genstring ()
  (symbol-name (gensym)))

(defmethod awake :before ((self auto-list))
  (with-slots (driver)
    self
    (setf driver
          (typecase driver
            (list
             (make-list-driver driver))
            (t driver)))))

(defmethod render-html ((self auto-list) stream)
  (with-slots (value current-value state driver any any-value none rows multiple element-method element-frame)
    self
    (setf state nil)
    (setf current-value value)
    (with-slots (data test key label-fn)
      driver
      (let ((options nil)
            (script-code ""))
        (flet ((write-option (symbol label selected?)
                 (if selected?
                     (html-stream
                      stream
                      ((:option :value symbol :selected "Y")
                       (:princ label)))
                     (html-stream
                      stream
                      ((:option :value symbol)
                       (:princ label)))))
               (add-option (symbol value label selected?)
                 (push (list symbol value label selected?) options)))
          (let ((data (etypecase data
                        (list data)
                        (function (funcall data)))))
            (when any
              (let ((anysym (genstring)))
                (push (cons anysym any-value) state)
                (add-option anysym t "Any" (equal value any-value))))
            (when none
              (let ((nonesym (genstring)))
                (push (cons nonesym nil) state)
                (add-option nonesym t "None" (null value))))
            (when (or any none)
              (let ((separator (genstring)))
                (push (cons separator separator) state)
                (add-option separator nil "  ----  " nil)))
            (dolist (row data)
              (let ((label (funcall label-fn row))
                    (option-value (funcall key row))
                    (itemsym (genstring)))
                (push (cons itemsym option-value) state)
                (add-option itemsym option-value label
                            (if multiple
                                (let ((selected? nil))
                                  (dolist (value (listify value))
                                    (if (funcall test value option-value)
                                        (setf selected? t)))
                                  selected?)
                                (funcall test value option-value)))))
            ;; generate script code
            (when element-method
              ;; generate the script code
              (let ((fn (make-string-output-stream)))
                (do ((case options (cdr case)))
                    ((null case))
                  (let ((case (car case)))
                    (destructuring-bind (symbol value label selected?)
                        case
                      (declare (ignore label selected?))
                      (if value
                          (format fn "if (popup_value(this) == '~a') {~%  form.target='~a';~%form.action='~a';~%form.submit();~%} "
                                  symbol
                                  (or element-frame "")
                                  (element-url (element-parent self)
					       :method (symbol-name element-method)
					       :args (or (element-method-args self) value)))
			(format fn "if (false) {} "))))
		  (if (cdr case)
                      (write-string "else " fn)))
                (setf script-code (get-output-stream-string fn))))
            ;; generate the tags
            (if multiple
                (html-stream
                 stream
                 ((:select :size rows
                           :multiple t
                           :onchange script-code
                           :name (element-external-name self))
                  (dolist (option (nreverse options))
                    (destructuring-bind (symbol value label selected?)
                        option
                      (declare (ignore value))
                      (write-option symbol label selected?)))))
                (html-stream
                 stream
                 ((:select :size rows
                           :onchange script-code
                           :name (element-external-name self))
                  (dolist (option (nreverse options))
                    (destructuring-bind (symbol value label selected?)
                        option
                      (declare (ignore value))
                      (write-option symbol label selected?))))))))))))

(defmethod accept-input :around ((self auto-list))
  (with-slots (primitive-value current-value state multiple)
    self
    (typecase primitive-value
      (cons
       (mapcar (lambda (value)
                 (cdr (assoc value state :test #'string=)))
               primitive-value))
      (atom
       (cdr (assoc primitive-value state :test #'string=))))))

(defclass list-driver ()
  ((data     :initarg :data)
   (test     :initarg :test)
   (key      :initarg :key)
   (label-fn :initarg :label)))

(defun make-list-driver (data &optional test key label)
  (make-instance 'list-driver
                 :data data
                 :test (or test #'equal)
                 :key (or key #'cadr)
                 :label (or label #'car)))

;; ------------------------------------------------------------
;; html-element: submit-button

(defclass submit-button (html-form)
  ())

(defmethod render-html ((self submit-button) stream)
  (with-slots (value element-external-name text)
    self
    (let* ((url (element-url (element-parent self)
			     :method (find-element-method self)
			     :args (element-method-args self)))
           (onclick (format nil "form.target = '~a'; form.action='~a'; form.submit();"
                            (element-frame self)
                            url)))
      (html-stream
       stream
       ((:input :type "submit"
                :onclick onclick
                :name element-external-name
                :value value))))))

;; ------------------------------------------------------------
;; html-element: file-upload

(defclass file-upload (html-form)
  ())

(defmethod render-html ((self file-upload) stream)
  (setf (element-value self) nil)
  (with-slots (element-external-name)
    self
    (html-stream stream ((:input :type :file :name element-external-name)))))

;; ------------------------------------------------------------
;; html-element: text-field

(defclass text-field (html-form-primitive)
  ((submit-on-return :initarg :submit-on-return
                     :initform nil)
   (visible          :initform t)
   (columns          :initarg :cols
                     :initform 30)))

(defmethod get-primitive-value ((self text-field))
  (or (element-value self) ""))

(defmethod accept-input :around ((self text-field))
  (slot-value self 'primitive-value))

(defmethod render-html ((self text-field) stream)
  (with-slots (element-external-name visible columns submit-on-return)
    self
    (html-stream
     stream
     ((:input :type (if visible "TEXT" "PASSWORD")
              :size (format nil "~d" columns)
              :onkeypress (format nil "if (CheckReturn(event)) { form.target='~a'; form.action='~a'; ~a} return true;"
                                  (element-frame self)
                                  (element-url (element-parent self)
                                               :method (find-element-method self)
                                               :args (element-method-args self))
                                  (if submit-on-return
                                      "form.submit(); "
                                      ""))
              :name element-external-name
              :value (get-primitive-value self))))))

;; ------------------------------------------------------------
;; html-element: text-area

(defclass text-area (html-form-primitive)
  ((wrap-type
    :initarg :wrap
    :initform :hard)
   (columns
    :initarg :cols
    :initform 40)
   (rows
    :initarg :rows
    :initform 5)
   (max-input-length
    :initform 7500)
   (hscroll
    :initarg :hscroll
    :initform nil)
   ))

(defmethod get-primitive-value ((self text-area))
  (let ((value (element-value self)))
    (typecase value
      (string    value)
      (function  (funcall value))
      (integer   (format nil "~d" value))
      (float     (format nil "~f" value))
      (t         ""))))

(defmethod accept-input :around ((self text-area))
  (with-slots (primitive-value max-input-length)
    self
    (cond ((or (not primitive-value)
               (= 0 (length primitive-value)))
           nil)
          ((< max-input-length (length primitive-value))
           (cmsg "Length of text: ~d: truncating" (length primitive-value))
           (subseq primitive-value 0 max-input-length))
          (t
           primitive-value))))

(defmethod render-html ((self text-area) stream)
  (with-slots (element-external-name rows columns wrap-type)
    self
    (html-stream
     stream
     ((:textarea :type :textarea
                 :wrap (case wrap-type
                         (:soft "SOFT")
                         (t     "HARD"))
                 :rows rows
                 :cols columns
                 :name element-external-name)
      (:princ (get-primitive-value self))))))

;; ------------------------------------------------------------
;; html-element: password-field

(defclass password-field (text-field)
  ())

(defmethod initialize-instance ((self password-field) &rest initargs)
  (declare (ignore initargs))
  (call-next-method)
  (setf (slot-value self 'visible) nil))
   
;; ------------------------------------------------------------
;; html-element: checkbox

(defclass checkbox (html-form-primitive)
  ())

(defmethod accept-input :around ((self checkbox))
  (with-slots (primitive-value)
    self
    (equal primitive-value "on")))

(defmethod render-html ((self checkbox) stream)
  (with-slots (element-external-name)
    self
    (format stream "<INPUT TYPE='CHECKBOX' NAME='~a' ~@[CHECKED~]>"
            element-external-name (element-value self))))

;; ------------------------------------------------------------
;; html-element: radio-button

(defclass radio-button (html-form)
  ((group   :initarg :group
            :initform :radio
            :documentation
            "Group to which a radio button belongs")
   (checked :initarg :checked
            :initform nil
            :documentation
            "Whether this button is initially checked in its group")))

;; A radio button returns t if it's on, nil otherwise

(defmethod element-value ((self radio-button))
  (let* ((gname          (slot-value self 'group))
         (self           (slot-value self 'element-external-name))
         (received-value (cadr
                          (assoc
                           gname (request-client-content *active-request*)
                           :test #'string-equal))))
    (string-equal received-value self)))

(defmethod render-html ((self radio-button) stream)
  (with-slots (element-external-name group checked)
    self
    (if checked
        (html-stream
         stream
         ((:input :type "radio"
                  :name group
                  :checked t
                  :value element-external-name)))
        (html-stream
         stream
         ((:input :type "radio"
                  :name group
                  :value element-external-name))))))

;; ------------------------------------------------------------
;; html-element: reset-button

(defclass reset-button (html-form)
  ())

(defmethod render-html ((self reset-button) stream)
  (html-stream
   stream
   ((:input :type :reset))))

;; ------------------------------------------------------------
;; html-element: file-chooser

(defclass file-chooser (html-form)
  ())

(defmethod render-html ((self file-chooser) stream)
  (html-stream
   stream
   ((:input :type :file))))
   
;; ------------------------------------------------------------
;; html-element: image-input

(defclass image-input (html-form image)
  ((source :initarg :source
           :initform nil)))
  
(defmethod render-html ((self image-input) stream)
  (with-slots (element-external-name source border)
    self
    (html-stream
     stream
     ((:input :type "image"
              :border border
              :name element-external-name
              :src source)))))

#+nil
(defmethod take-values-from-request ((self image-input))
  (setf (element-value self) nil)
  (let ((valuex (cadr (assoc (concatenate 'string
                                          (element-external-name self)
                                          ".x")
                             (request-client-content request)
                             :test #'equal))))
    (if valuex
        (setf (element-value self) t))))

;; ------------------------------------------------------------
;; html-element: button

(defclass button (html-form enabling-mixin)
  ()
  (:documentation
   "An HTML form button"))

(defmethod set-element-value ((self button) value)
  (when (stringp value)
    (setf (slot-value self 'value) value)))

(defmethod render-html ((self button) stream)
  (with-slots (element-external-name text);; THINKME: enabled
    self
    (let ((onclick (format nil "form.target = '~a'; form.action='~a'; form.submit();"
                           ;; an actionless button is simply process-form-submit
                           (element-frame self)
                           (element-url (element-parent self)
					:method (find-element-method self)
					:args (element-method-args self)))))
      (html-stream
       stream
       ((:input :type "button"
                :onclick onclick
                :name element-external-name
                :value (element-value self)))))))

;;; repeater functions
;;;
;;; Any html-form element can add subelements at page render time
;;; using repeat-element, and look up these elements' values at method
;;; invocation time, using repeat-value

;; repeat-element
;;
;; Subelements added via repeat-element are removed after their values
;; are taken.  (see 'dynamic?' argument to make-child)

(defmethod repeat-element ((self html-form) stream &key name (dynamic t) reference-value class value initargs)
  (assert class (class) "Class not specified to repeat-element")
  (unless name
    (setf name (gensym)
          dynamic t))
  (let ((child (apply #'make-child `(,self ,class ,name ,dynamic :reference-value ,reference-value ,@initargs))))
    (setf (element-value child) value)
    (preawake child)
    (awake child)
    (render-html child stream)))

;; If reference-value and reference-test are supplied, make sure
;; those match too.
;;
;; If toggle is t, return the reference-values of subelements which
;; match NAME when the subelement's form value is not null. (good for
;; linking checkboxes to internal objects)
;;
;; If :return is :values, return the values of the requested elements
;; (the default); otherwise, return the elements themselves.

(defmethod repeat-value ((self html-form) &key (return :values) name toggle reference-value reference-test &aux result)
  (maphash (lambda (child-name child-list)
             (dolist (child child-list)
               (let* ((child-element (car child))
                      (child-value (element-value child-element))
		      (child-ref-value (element-reference-value child-element)))
                 (if (null name)
                     (push child-element result)
                     (when (equal name child-name)
                       (if toggle
                           ;; the element is a toggle
                           (when child-value
                             (push (element-reference-value child-element) result))
                           (when (or (null reference-value)
                                     (funcall reference-test child-ref-value reference-value))
                             (push child-element result))))))))
           (element-children self))
  (when (and (eql return :values)
             (not toggle))
    (setf result (mapcar #'element-value result)))
  (nreverse result))
