;;; -*- Syntax: Ansi-Common-Lisp; Base: 10; Mode: lisp; Package: imho -*-
;;; $Id: html-elements.lisp,v 1.80 2003/03/26 19:52:16 adam Exp $
;;;
;;; Copyright (c) 1999 - 2003 onShore Development, Inc.
;;;
;;; See the file 'COPYING' in this directory for terms.
;;;
;;; Basic non-form elements

(in-package :imho)

(eval-when (:compile-toplevel :load-toplevel)

(defmacro render-child (element child stream)
  `(render-html (child-element ,element ,child) ,stream))
  
(defun local-href (url)
  (concatenate 'string (application-doc-root *active-application*) url))

(declaim (inline local-href))

(defun write-spacer (stream &key (height 1) (width 1))
  (html-stream
   stream
   ((:img :src (local-href "images/spacer.gif") :width width :height height :alt ""))))

;; ------------------------------------------------------------
;; html-element: html-page

(defclass html-page (html-element)
  ())

(defclass html-frameset (html-element)
  ((title        :initarg :title
                 :initform "")
   (cols         :initarg :cols
                 :initform nil)
   (rows         :initarg :rows
                 :initform nil)
   (border       :initarg :border
                 :initform 0)
   (framespacing :initarg :framespacing
                 :initform 0)
   (frameborder  :initarg :frameborder
                 :initform 0)
   (frames       :initarg :frames
                 :initform nil
                 :documentation "The (internal) names of frames making up this set, in order.")))

(defmethod render-html ((self html-frameset) stream)
  (with-slots (title cols rows frames border framespacing frameborder)
    self
    (set-page-title title)
    (flet ((render-frames ()
             (dolist (frame frames)
               (destructuring-bind (name method &optional (scrolling :auto))
                   frame
                 (declare (ignore method))
                 (let ((frame-url (build-element-url self :reference (refer-wm load-frame (symbol-name name)))))
                   (html-stream
                    stream
                    ((:frame :src frame-url :name name :scrolling (symbol-name scrolling)))))))))
      (if rows
          (html-stream
           stream
           ((:frameset :title title :rows rows :border border :framespacing framespacing :frameborder frameborder)
            (render-frames)))
          (html-stream
           stream
           ((:frameset :title title :cols cols :border border :framespacing framespacing :frameborder frameborder)
            (render-frames)))))))

;; load the contents of a frame (one element of a frameset).
;;
;; If the frame's content already exists, assume that preawak/awake
;; has already been called on it, since load-frame is only called as a
;; consequence of frameset load.
;;
;; If the content element doesn't exist, generate it by calling the
;; default method of the frame element, add the 'child', and request
;; preawake/awake.

(define-wm load-frame ((self html-frameset) (frame keyword))
  (if-bind (child (child-element self frame :if-does-not-exist :ignore))
      (values child t)
      (let ((method (cadr (assoc frame (slot-value self 'frames)))))
        (unless method
          (error "No method defined for frame ~s" frame))
        (setf (child-element self frame nil) (apply-wm method)))))

;; ------------------------------------------------------------

(defclass sized-mixin ()
  ((width
    :initarg :width
    :initform 0)
   (height
    :initarg :height
    :initform 0)))

(defclass enabling-mixin ()
  ((enabled
    :initarg :enabled
    :accessor element-enabled
    :initform t)))


;; ------------------------------------------------------------
;; html-element: hyperlink
;;
;; A text hyperlink.

(defclass hyperlink (method-caller enabling-mixin html-element)
  ())

;; ------------------------------------------------------------
;; framework-method: element-value
;;
;; Allow the 'value' slot of a hyperlink to be a zero-arity function
;; that returns the content for the link.

(defmethod element-value ((element hyperlink))
  (with-slots (value)
    element
    (typecase value
      (function (funcall value))
      (string value)
      (null "")
      (t value))))

;; By default, spaces inside of hyperlinks are converted to &nbsp;'s

(defmethod render-html ((self hyperlink) stream)
  (let ((link-text (string-replace #\Space "&nbsp;" (element-value self))))
    (if (element-enabled self)
        (html-stream
         stream
         ((:a :href (element-url self)
              :target (element-frame self))
          (:princ link-text)))
        (write-string link-text stream))))

(defclass tooltip-mixin ()
  ((tooltip
    :initarg :tooltip
    :initform nil)))

(defclass status-msg-mixin ()
  ((status-msg
    :initarg :status-msg
    :initform nil)))

(defclass image (sized-mixin tooltip-mixin status-msg-mixin enabling-mixin html-element)
  ((source
    :initarg :source
    :initform nil)
   (alttext
    :initarg :alt
    :initform nil)
   (border
    :initarg :border
    :initform 0)
   (selected :initform nil)))

(defmethod render-html ((element image) stream)
  (with-slots (alttext source selected width height enabled border tooltip element-external-name)
    element
    (let ((image-src (get-image-url source (if enabled (if selected :clicked :default) :disabled))))
      (if (and width height)
          (html-stream
           stream
           ((:img :src image-src
                  :alt (if (and alttext (not tooltip)) alttext "")
                  :width width
                  :height height
                  :name element-external-name
                  :border 0)))
          (html-stream
           stream
           ((:img :src image-src
                  :alt (if (and alttext (not tooltip)) alttext "")
                  ;; :width (format nil "~d" width)
                  ;; :height (format nil "~d" height)
                  :name element-external-name
                  :border 0)))))))

;; ------------------------------------------------------------
;; image-button

(defclass image-button (image method-caller)
  ())

(defmethod render-html ((element image-button) stream)
  (with-slots (source width height enabled selected tooltip status-msg element-external-name element-method-args)
    element
    (cond ((or selected (not enabled))
           (call-next-method))
          (t
           (let ((standard (get-image-url source :default))
                 (highlit  (get-image-url source :highlit))
                 (clicked  (get-image-url source :clicked))
                 (tooltip-out "")
                 (tooltip-over "")
                 (status-over "")
                 (status-out ""))
             (scripted-element-init "IMHOLoadImage" element-external-name standard highlit clicked "")
             (when tooltip
               (let ((divname (concatenate 'string element-external-name "DIV")))
                 (setq tooltip-over (format nil "doToolTip(event, '~a'); "
                                            divname))
                 (setq tooltip-out (format nil "clearToolTip('~a'); "
                                           divname))
                 (write-layer divname
                              '(("position" . "absolute")
                                ("visibility" . "hidden")
                                ("margin-left" . "250 px")
                                ("margin-right" . "250 px")
                                ("margin-top" . "25")
                                ("z-index" . "1")
                                ("text-align" . "center")
                                ("font-family" . "arial, helvetica")
                                ("font-size" . "12pt")
                                ("border-width" . "0")
                                ("background-color" . "e4e4e4")
                                ("text-color" . "000033")))
                 (with-div (divstream divname)
                   (html-stream
                    divstream
                    ((:table :width "300" :height "28" :border 0)
                     (:tr
                      ((:td :align :center :class "flabel_2")
                       (write-string tooltip divstream))))))))
             (when status-msg
               (setq status-over (format nil "doStatusMsg('~a'); "
                                         status-msg))
               (setq status-out (format nil "clearStatusMsg(); ")))
             (let ((frame-args (element-frame-args element))
                   (url (element-url (element-parent element) :method (find-element-method element)
                                     :args element-method-args)))
               (if frame-args
                   (html-stream
                    stream
                    ((:a :href "#"
                         :onclick
                         (format nil "window.open('~a',~a)" url frame-args)
                         :target (element-frame element)
                         :onmouseover (format nil "~a; ~areturn CSIShow('~a',1);"
                                              status-over tooltip-over element-external-name)
                         :onmouseout (format nil "~a; ~areturn CSIShow('~a',0);"
                                             status-out tooltip-out element-external-name)
                         :onmousedown (format nil "CSIShow('~a',2); return CSButtonReturn()"
                                              element-external-name))
                     (call-next-method)))
                   (html-stream
                    stream
                    ((:a :href url
                         :target (element-frame element)
                         :onmouseover (format nil "~a; ~areturn CSIShow('~a',1);"
                                              status-over tooltip-over element-external-name)
                         :onmouseout (format nil "~a; ~areturn CSIShow('~a',0);"
                                             status-over tooltip-out element-external-name)
                         :onmousedown (format nil "CSIShow('~a',2); return CSButtonReturn()"
                                              element-external-name))
                     (call-next-method))))))))))

(defclass splash-button (image method-caller)
  ())

(defmethod render-html ((element splash-button) stream)
  (with-slots (source status-msg width height enabled element-external-name)
     element
     (let ((status-over "")
	   (status-out ""))
       (when status-msg
	 (setq status-over (format nil "doStatusMsg('~a'); "
				   status-msg))
	 (setq status-out (format nil "clearStatusMsg(); ")))
       (scripted-element-init "CSNullInit" element-external-name "splash.html" "/splash.html" 400 400)
       (html-stream
	stream
	((:a :href (element-url element)
	     :onmouseover status-over
	     :onmouseout status-out
	     :onclick (format nil "DOSplash(); return ! CSButtonReturn();"
			      element-external-name))
	 (call-next-method))))))

)
