;;; -*- Syntax: Ansi-Common-Lisp; Base: 10; Mode: lisp; Package: imho -*-
;;; $Id: html-presentation.lisp,v 1.10 2003/03/26 19:52:16 adam Exp $
;;;
;;; Copyright (c) 1999 - 2003 onShore Development, Inc.
;;;
;;; See the file 'COPYING' in this directory for terms.

(in-package :imho)

;; presentation types can define an accept editor which returns a set
;; of arguments for making an instead of a form-element, which will be
;; passed to repeat-element in the default accept method for html-view
(define-presentation-generic-function %accept-editor accept-editor (type-key parameters type object query-identifier))


(defun accept-editor (object type query-identifier)
  (funcall-presentation-generic-function accept-editor object type query-identifier))

(define-presentation-method accept-editor ((type t) value property)
  (declare (ignore value property))
  (with-presentation-type-decoded (name parameters)
    type
    (declare (ignore parameters))
    name))

(defmethod repeat-element ((self html-view) stream &key name (dynamic t) reference-value class value initargs)
  (assert class (class) "Class not specified to repeat-element")
  (unless name
    (setf name (gensym)
          dynamic t))
  (let ((child (apply #'make-instance `(,class :reference-value ,reference-value ,@initargs))))
    (setf (element-value child) value)
    (preawake child)
    (awake child)
    (render-html child stream)))


(defmethod edited-property-value ((self html-view) property)
  (declare (ignore property))
  (values nil nil))

(defmethod html-accept  ((type t) stream (view html-view)
                                    &key (default nil defaultp)
                                    (insert-default t)
                                    (query-identifier (gensym))
                                    (prompt nil))
  (declare (ignore prompt))
  (destructuring-bind (editor-class &rest initargs)
      (listify (accept-editor type default query-identifier))
    (if editor-class
        (repeat-element view stream
                        :name query-identifier
                        :reference-value (edited-property-value view query-identifier)
                        :value (when (and defaultp insert-default) default)
                        :class editor-class
                        :initargs initargs)
        (format stream "[EDIT ~s / ~s]" type default))))


;; presentation methods for base presentation types.  Many present
;; methods are just fine with the textual-view implementation.  We
;; need to override accept methods for all base types since they will
;; specialize on the type and not the view first

;; presentation methods for html-view

(define-presentation-method accept ((type t) stream (view html-view)
                                    &rest rest-args)
  (apply #'html-accept type stream view rest-args))


(define-presentation-method present (object (type null)
                                            stream
				     (view html-view)
				     &key acceptably for-context-type)
  (declare (ignore object acceptably for-context-type))
  (write-string "<i>None<i>" stream))

;; boolean
(define-presentation-method accept-editor ((type boolean) value property)
  (declare (ignore value property))
  'checkbox)

(define-presentation-method accept ((type boolean) stream (view html-view)
                                    &rest rest-args)
  (apply #'html-accept type stream view rest-args))

;; symbol
(define-presentation-method present (object (type symbol) stream
				     (view html-view)
				     &key acceptably for-context-type)
  (declare (ignore acceptably for-context-type))
  (if object
      (write-string (string-capitalize (symbol-name object)) stream)
      (write-string "<i>None</i>" stream)))

(define-presentation-method accept-editor ((type symbol) value property)
  (declare (ignore value property))
  (list 'text-field :cols 40))

(define-presentation-method accept ((type symbol) stream (view html-view)
                                    &rest rest-args)
  (apply #'html-accept type stream view rest-args))

;; keyword
(define-presentation-method present (object (type keyword) stream
				     (view html-view)
				     &key acceptably for-context-type)
  (declare (ignore acceptably for-context-type))
  (if object
      (write-string (string-capitalize (symbol-name object)) stream)
      (write-string "<i>None</i>" stream)))

(define-presentation-method accept ((type keyword) stream (view html-view)
                                    &rest rest-args)
  (apply #'html-accept type stream view rest-args))


;; numbers

(define-presentation-method accept-editor ((type real) value property)
  (declare (ignore value property))
  (list 'text-field :cols 10))

(define-presentation-method accept ((type real) stream (view html-view)
                                    &rest rest-args)
  (apply #'html-accept type stream view rest-args))

;; integer

(define-presentation-type integer (&optional low high) :options ((base 10) radix)
			  :inherit-from 'number)

(define-presentation-method present (value (type real) stream
                                           (view html-view)
                                           &key acceptably for-context-type)
  (declare (ignore acceptably for-context-type))
  (if value
      (let ((*print-base* base)
            (*print-radix* radix))
        (princ value stream))
      (write-string "<i>None</i>" stream)))
  
(define-presentation-method accept-editor ((type real) value property)
  (declare (ignore value property))
  (list 'text-field :cols 10))



;; string
(define-presentation-method present (value (type string) stream
                                           (view html-view)
                                           &key acceptably for-context-type)
  (declare (ignore acceptably for-context-type))
  (if (null value)
      (write-string "<i>None</i>" stream)
      (if (string-empty-p value)
          (write-space stream)
          (write-string value stream))))

(define-presentation-method accept-editor ((type string) value property)
  (declare (ignore value property))
  (list 'text-field :cols (or (and (numberp length) length)
                              (and (numberp max-length) max-length)
                              40)))

(define-presentation-method accept ((type string) stream (view html-view)
                                    &rest rest-args)
  (apply #'html-accept type stream view rest-args))

(define-presentation-method accept ((type t) stream (view html-view)
                                    &key (default nil defaultp)
                                    (insert-default t)
                                    (query-identifier (gensym))
                                    (prompt nil)
                                    &allow-other-keys
                                    &rest rest-args)
  (declare (ignore defaultp insert-default prompt))
  (if (accept-editor type default query-identifier)
      (apply #'html-accept type stream view rest-args)
      (format stream "No accept-editor or html-view accept  method defined for ~A" type)))


(defmacro define-html-accept (type)
  `(define-presentation-method accept ((type ,type) stream (view html-view)
                                       &rest rest-args)
    (apply #'html-accept type stream view rest-args)))

(define-presentation-method present ((object null) (type sequence)
				     stream
				     (view html-view)
				     &key acceptably for-context-type)
  (declare (ignore acceptably for-context-type))
  (write-string "<i>None</i>" stream))


(define-presentation-method present ((object null) (type t)
				     stream
				     (view html-view)
				     &key acceptably for-context-type)
  (declare (ignore acceptably for-context-type))
  (write-string "<i>None</i>" stream))

